package com.sillycycle.bagleyd.welltris;

/*
 * @(#)WelltrisCanvas.java
 *
 * Copyright 1993 - 2020  David A. Bagley, bagleyd AT verizon.net
 *
 * Taken from GENERIC X-BASED TETRIS
 *
 * Copyright 1992 - 1995 Q. Alex Zhao, azhao AT cc.gatech.edu
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Point;
import java.util.Random;
import javax.swing.JFrame;

/**
 * The <code>WelltrisCanvas</code> class does all the graphics for
 * <code>WelltrisFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/altris.html
 */

public class WelltrisCanvas extends Canvas {
	private static final long serialVersionUID = 42L;

	transient PolyMode[] tris; //[PolyominoThing.MAX_UNITS - PolyominoThing.MIN_UNITS + 1];
	static int[] frozenWall = new int[WelltrisInterface.MAX_SIDES];

	Random generator;
	int startlevel = WelltrisInterface.DEFAULT_LEVEL;
	int level = WelltrisInterface.DEFAULT_LEVEL;
	int diameter = WelltrisInterface.DEFAULT_WIDTH;
	int depth = WelltrisInterface.DEFAULT_DEPTH;
	int fill = WelltrisInterface.DEFAULT_FILL;
	boolean fixedFill = false, suddenAppear = false;
	boolean bonus = false;
	boolean cw = true, grid = true, showNext = false;
	boolean invert = false, dropPaused = false;
	boolean firstPaint = false, resizePaint = false;
	boolean redrawPaint = false, framePaint = false;
	boolean focus = true;
	int score = 0, completedRows = 0;
	int perimeter = diameter * WelltrisInterface.MAX_SIDES;
	int boxSize = WelltrisInterface.DEFAULT_BOXSIZE;
	int boxWidth = boxSize, boxHeight = boxSize;
	int viewLength = 3 * boxSize * diameter + 2;
	Point fieldOffset = new Point(2, 2);
	int delta = 1;
	int baseWidth = diameter * boxSize + delta;
	int endPt = 3 * baseWidth;
	Color background;
	transient Thing curThing, nextThing;
	transient Spec curSpec = new Spec(PolyominoThing.DEFAULT_UNITS);
	transient Spec nextSpec = new Spec(PolyominoThing.DEFAULT_UNITS);
	PolyominoImage polyominoImage;
	transient Field[][] wall = null;
	transient Field[][] base = null;
	JFrame frame = null;
	boolean debug = false;

	public WelltrisCanvas() {
		curThing = new Thing();
		nextThing = new Thing();
		polyominoImage = new PolyominoImage();
		generator = new Random(System.nanoTime());
		readPolyominoes();
		sizeInit();
		inits();
	}

	void specInit(int units, boolean diagonal, boolean mixed) {
		curSpec = new Spec(units, diagonal, mixed, false);
		nextSpec = new Spec(units, diagonal, mixed, false);
	}

	final void sizeInit() {
		wall = new Field[depth + diameter][perimeter];
		base = new Field[diameter][diameter];
		for (int z = 0; z < depth + diameter; z++)
			for (int x = 0; x < perimeter; x++)
				wall[z][x] = new Field();
		for (int y = 0; y < diameter; y++)
			for (int x = 0; x < diameter; x++)
				base[y][x] = new Field();
	}

	final void inits() {
		int i, j, turnStyle, numberPolyominoes;

		turnStyle = tris[nextSpec.units - PolyominoThing.MIN_UNITS].mode
			[(nextSpec.diagonal) ? 1 : 0].turnStyle;
		if (turnStyle < 0 || turnStyle >= PolyominoThing.MAX_TURN_TYPES) {
			System.err.println("Welltris: corrupted polyomino file");
			System.err.println("turnStyle = " + turnStyle);
		}
		numberPolyominoes = tris[nextSpec.units - PolyominoThing.MIN_UNITS].mode
			[(nextSpec.diagonal) ? 1 : 0].number[(nextSpec.mixed) ? 1 : 0][turnStyle];
		if (numberPolyominoes <= 0 ||
				numberPolyominoes > PolyominoThing.MAX_START_POLYOMINOES) {
			System.err.println("Welltris: corrupted polyomino file");
			System.err.println("\tnumberPolyominoes = " + numberPolyominoes);
		}
		for (j = 0; j < depth + diameter; j++)
			for (i = 0; i < perimeter; i++) {
				wall[j][i].pmid = WelltrisInterface.EMPTY;
				wall[j][i].cid = 0;
			}
		for (j = 0; j < diameter; j++)
			for (i = 0; i < diameter; i++) {
				base[j][i].pmid = WelltrisInterface.EMPTY;
				base[j][i].cid = 0;
			}
		unfreezeWalls();
		initFill();
		/* new things */
		newThing();
		newThing();
	}

	Point wallToBase(int wallX, int wallY) {
		switch (wallX / diameter) {
		case 0:
			return new Point(wallX, wallY - depth);
		case 1:
			return new Point (diameter - 1 + depth - wallY,
					wallX - diameter);
		case 2:
			return new Point(perimeter - 1 - diameter - wallX,
					diameter - 1 + depth - wallY);
		case 3:
			return new Point(wallY - depth,
					perimeter - 1 - wallX);
		default:
			System.err.println("wallToBase kinds range 0-3, got " +
				wallX / diameter);
			return new Point(-1, -1);
		}
	}

	int checkWall(int pmid, int baseI, int baseJ) {
		int newPmid = pmid;

		if (newPmid < 0)
			return newPmid;
		/* Check if next to well wall */
		if (baseI == 0)
			newPmid = PolyominoImage.CHECKLEFT(newPmid);
		if (baseJ == 0)
			newPmid = PolyominoImage.CHECKUP(newPmid);
		if (baseI == diameter - 1)
			newPmid = PolyominoImage.CHECKRIGHT(newPmid);
		if (baseJ == diameter - 1)
			newPmid = PolyominoImage.CHECKDOWN(newPmid);
		return newPmid;
	}

	int checkSides(int pmid, int i) {
		int newPmid = pmid;

		if (newPmid < 0)
			return newPmid;
		if (i == 0 || i == 3 * diameter - 1)
			newPmid = PolyominoImage.CHECKLEFT(newPmid);
		else if (i == 2 * diameter || i == diameter - 1)
			newPmid = PolyominoImage.CHECKRIGHT(newPmid);
		else if (i ==  diameter || i == 4 * diameter - 1)
			newPmid = PolyominoImage.CHECKUP(newPmid);
		else if (i == 3 * diameter || i == 2 * diameter - 1)
			newPmid = PolyominoImage.CHECKDOWN(newPmid);
		return newPmid;
	}

	int checkBottom(int pmid, int i, int j) {
		if (pmid < 0 || j != depth - 1)
			return pmid;
		switch (i / diameter) {
		case 0:
			return PolyominoImage.CHECKDOWN(pmid);
		case 1:
			return PolyominoImage.CHECKLEFT(pmid);
		case 2:
			return PolyominoImage.CHECKUP(pmid);
		case 3:
			return PolyominoImage.CHECKRIGHT(pmid);
		default:
		}
		return WelltrisInterface.EMPTY;
	}

	int checkWallSides(int pmid, int i) {
		int newPmid = pmid;

		if (newPmid < 0)
			return newPmid;
		if (i % diameter == 0)
			newPmid = PolyominoImage.CHECKLEFT(newPmid);
		else if (i % diameter == diameter - 1)
			newPmid = PolyominoImage.CHECKRIGHT(newPmid);
		return newPmid;
	}

	int checkWallBottom(int pmid, int j) {
		if (pmid < 0 || j != depth - 1)
			return pmid;
		return PolyominoImage.CHECKDOWN(pmid);
	}

	int rotateIndex(int pmid, int i, int j) {
		int wallRotation, pm1, pm0;

		if (pmid == WelltrisInterface.EMPTY)
			return WelltrisInterface.EMPTY;
		if (j < depth)
			return pmid;
		wallRotation = (WelltrisInterface.MAX_SIDES - i /
			diameter) % WelltrisInterface.MAX_SIDES;
		pm1 = (0xF0 & pmid) << wallRotation;
		pm0 = (0xF & pmid) << wallRotation;
		pm1 = (0xF0 & pm1) | ((pm1 & 0xF00) >> WelltrisInterface.MAX_SIDES);
		pm0 = (0xF & pm0) | ((pm0 & 0xF0) >> WelltrisInterface.MAX_SIDES);
		return (pm1 | pm0);
	}

	void newThing() {
		int w;

		curThing.clone(nextThing);
		curSpec.clone(nextSpec);
		nextThing.setRandomNumber(Math.abs(generator.nextInt()));
		nextThing.setRotation(generator.nextInt(WelltrisInterface.MAX_SIDES));
		nextThing.setReflection(generator.nextInt(2));
		nextThing.setBonus(nextSpec.bonus);
		curThing.position.x = generator.nextInt(diameter - curThing.size + 1);
		/* curThing.position.x = (diameter - curThing.size) / 2; */
		/* not all frozen else this is an infinite loop */
		do {
			w = generator.nextInt(WelltrisInterface.MAX_SIDES);
		} while (frozenWall[w] != 0);
		curThing.position.x += w * diameter;
		redoNext();
	}

	void drawTitle() {
		//WelltrisApp.drawTitle
	}

	void drawStatus() {
		/*
		WelltrisApp.drawScore("Score: " + score);
		WelltrisApp.drawLevel("Level: " + level);
		WelltrisApp.drawRows("Rows: " + complededRows);
		*/
	}

	void clearField() {
		Graphics g = this.getGraphics();

		// This will flash
		PolyominoImage.clearUnits(g, background, 0, 0, 0, 0,
			getSize().width, getSize().height);
		boxWidth = getSize().width;
		boxHeight = getSize().height;
		if (boxWidth > boxHeight)
			boxSize = boxHeight;
		else
			boxSize = boxWidth;
		boxSize = (boxSize - 8) / (3 * diameter);
		boxWidth = boxSize;
		boxHeight = boxSize;
		viewLength = 3 * boxSize * diameter + 2;
		fieldOffset = new Point((getSize().width - 2 - viewLength) / 2,
			(getSize().height - 2 - viewLength) / 2);
		baseWidth = diameter * boxSize + delta;
		endPt = 3 * baseWidth;
		resetImage();
		/* for (int i = 0; i < perimeter; i++)
			for (int k = 0; k < depth; k++)
				drawTrapazoid(g, background,
					//wall[k][i].pmid
					i, k);
		for (int i = 0; i < diameter; i++)
			for (int j = 0; j < diameter; j++)
				clearUnit(g, background, i, j); */
	}

	void resetImage() {
		if (this.boxWidth > 2) {
			polyominoImage.createBitmaps(false, false,
				this.boxWidth, background);
			polyominoImage.createBitmaps(true, false,
				this.boxWidth, background);
		}
	}

	void draw3DFrame(boolean inside, int x, int y, int width, int height) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			if (inside)
				g.setColor(background.darker());
			else
				g.setColor(background);
			g.fillRect(x, y, width, 1);
			g.fillRect(x, y + 1, 1, height - 1);
			g.fillRect(x + 1, y + 1, width - 2, 1);
			g.fillRect(x + 1, y + 2, 1, height - 3);
			if (inside)
				g.setColor(background.brighter().brighter());
			else
				g.setColor(background);
			g.fillRect(x + 1, y + height - 1, width - 1, 1);
			g.fillRect(x + width - 1, y + 1, 1, height - 2);
			g.fillRect(x + 2, y + height - 2, width - 3, 1);
			g.fillRect(x + width - 2, y + 2, 1, height - 4);
		} finally {
			g.dispose();
		}
	}

	void drawFrame(Graphics g, boolean show) {
		int offsetX = fieldOffset.x - 2;
		int offsetY = fieldOffset.y - 2;
		int lengthX = fieldOffset.x + viewLength + 4;
		int lengthY = fieldOffset.y + viewLength + 4;

		if (g == null)
			return;
		g.setColor(background.darker().darker());
		g.fillRect(0, 0, getSize().width, offsetY);
		g.fillRect(0, offsetY, offsetX, lengthY - offsetY);
		g.fillRect(0, lengthY, getSize().width, offsetY);
		g.fillRect(lengthX, offsetY,
			getSize().width - lengthX, lengthY - offsetY);
		draw3DFrame(show && focus, offsetX, offsetY,
			lengthX - offsetX, lengthY - offsetY);
	}

	void drawFrame(boolean show) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			drawFrame(g, show);
		} finally {
			g.dispose();
		}
	}

	void drawField() {
		Graphics g = this.getGraphics();

		for (int j = 0; j < depth; j++) {
			for (int i = 0; i < perimeter; i++) {
				if (wall[j][i].pmid >= 0) {
					drawTrapazoid(g,
						wall[j][i].pmid,
						PolyominoImage.colorThings[wall[j][i].cid],
						((grid) ? Color.black : background),
						i, j, PolyominoImage.DRAW_COLOR);
				}
				if (frozenWall[i / diameter] != 0) { // Simulate Xor
					Color color;
					if (wall[j][i].pmid < 0)
						color = background;
					else
						color = PolyominoImage.colorThings[wall[j][i].cid];
					color = new Color(255 - color.getRed(),
						255 - color.getGreen(), 255 - color.getBlue());
					drawTrapazoid(g,
						wall[j][i].pmid,
						color,
						((grid) ? Color.black : background),
						i, j,
						((wall[j][i].pmid >= 0) ?
						PolyominoImage.XOR_COLOR : PolyominoImage.XOR_CLEAR_COLOR));
				}
			}
		}
		for (int i = 0; i < diameter; i++)
			for (int j = 0; j < diameter; j++)
				if (base[j][i].pmid >= 0) {
				  drawUnit(g, base[j][i].pmid, base[j][i].cid,
				    i, j);
				}
	}

	void drawThing() {
		Graphics g = this.getGraphics();
		int i, j, xi, yj;
		int[][] tempBase = new int[diameter][diameter];

		/* Initialize */
		for (i = 0; i < diameter; i++)
			for (j = 0; j < diameter; j++)
				tempBase[j][i] = WelltrisInterface.EMPTY;
		/* Find pieces and overlap them if necessary */
		for (i = 0; i < curThing.size; i++)
			for (j = 0; j < curThing.size; j++)
				if (tris[curThing.unitsIndex].poly
				      [curThing.polyNumber].shape[j][i] != 0) {
				  xi = (curThing.position.x + i) % perimeter;
				  yj = curThing.position.y + j;
				  if (yj >= depth) {
				    int rotPmid = rotateIndex(
				      PolyominoImage.polyPixmap(i, j,
				      tris[curThing.unitsIndex].poly[curThing.polyNumber].shape,
				      curThing.size),
				      xi, yj);

				    Point baseCoord = wallToBase(xi, yj);
				    xi = checkWall(rotPmid, baseCoord.x, baseCoord.y);
				    tempBase[baseCoord.y][baseCoord.x] &= xi;
				  }
				}
		/* Now draw */
		for (i = 0; i < curThing.size; i++)
			for (j = 0; j < curThing.size; j++)
				if (tris[curThing.unitsIndex].poly
				      [curThing.polyNumber].shape[j][i] != 0) {
				  xi = (curThing.position.x + i) % perimeter;
				  yj = curThing.position.y + j;
				  if (yj >= depth) {
				    Point baseCoord = wallToBase(xi, yj);
				    if (tempBase[baseCoord.y][baseCoord.x] != WelltrisInterface.EMPTY) {
				      /* drawUnit(g, polyominoImage.polyPixmap(i, j,
					curThing, tempBase),
					curThing.colorNumber,
					baseCoord.x, baseCoord.y); */
				      /* This uses precomputed values */

				      drawUnit(g,
					tempBase[baseCoord.y][baseCoord.x],
					curThing.colorNumber,
					baseCoord.x, baseCoord.y);
				      tempBase[baseCoord.y][baseCoord.x] = WelltrisInterface.EMPTY;
				    }
				  } else if (yj >= 0) {
				    int rotPmid = rotateIndex(
				      PolyominoImage.polyPixmap(i, j,
				      tris[curThing.unitsIndex].poly[curThing.polyNumber].shape,
				      curThing.size),
				      xi, yj);

				    drawTrapazoid(g,
				      checkWallSides(checkWallBottom(rotPmid, yj), xi),
				      PolyominoImage.colorThings[curThing.colorNumber],
				      ((grid) ? Color.black : background),
				      xi, yj, PolyominoImage.DRAW_COLOR);
				  }
				}
	}

	void drawThingDiff(Thing oldThing) {
		Graphics g = this.getGraphics();
		int i, j, xi, yj, ox, oy;
		int[][] tempBase = new int[diameter][diameter];

		/* Initialize */
		for (i = 0; i < diameter; i++)
			for (j = 0; j < diameter; j++)
				tempBase[j][i] = WelltrisInterface.EMPTY;
		/* Find pieces and overlap them if necessary */
		for (i = 0; i < curThing.size; i++)
			for (j = 0; j < curThing.size; j++)
				if (tris[curThing.unitsIndex].poly
				      [curThing.polyNumber].shape[j][i] != 0) {
				  xi = (curThing.position.x + i) % perimeter;
				  yj = curThing.position.y + j;
				  if (yj >= depth) {
				    int rotPmid = rotateIndex(
				      PolyominoImage.polyPixmap(i, j,
				      tris[curThing.unitsIndex].poly[curThing.polyNumber].shape,
				      curThing.size),
				      xi, yj);

				    Point baseCoord = wallToBase(xi, yj);
				    xi = checkWall(rotPmid, baseCoord.x, baseCoord.y);
				    tempBase[baseCoord.y][baseCoord.x] &= xi;
				  }
				}
		/* Test to make corner split neighbor awareness */
		for (i = 0; i < diameter; i++)
			for (j = 0; j < diameter; j++) {
			  if (tempBase[j][i] >= 0) {
			    if ((j > 0) && (tempBase[j - 1][i] != WelltrisInterface.EMPTY)) {
			      tempBase[j][i] &= ~PolyominoImage.UP;
			    }
			    if ((j < diameter - 1) &&
				(tempBase[j + 1][i] != WelltrisInterface.EMPTY)) {
			      tempBase[j][i] &= ~PolyominoImage.DOWN;
			    }
			    if (i > 0 && (tempBase[j][i - 1] != WelltrisInterface.EMPTY)) {
			      tempBase[j][i] &= ~PolyominoImage.LEFT;
			    }
			    if ((i < diameter - 1) &&
				(tempBase[j][i + 1] != WelltrisInterface.EMPTY)) {
			      tempBase[j][i] &= ~PolyominoImage.RIGHT;
			    }
			    if (i > 0 && j > 0 && (tempBase[j - 1][i - 1] != WelltrisInterface.EMPTY)) {
			      tempBase[j][i] &= ~PolyominoImage.LEFT_UP;
			    }
			    if (i > 0 && (j < diameter - 1) &&
				(tempBase[j + 1][i - 1] != WelltrisInterface.EMPTY)) {
			      tempBase[j][i] &= ~PolyominoImage.LEFT_DOWN;
			    }
			    if ((i < diameter - 1) && j > 0 &&
				(tempBase[j - 1][i + 1] != WelltrisInterface.EMPTY)) {
			      tempBase[j][i] &= ~PolyominoImage.RIGHT_UP;
			    }
			    if ((i < diameter - 1) && (j < diameter - 1) &&
				(tempBase[j + 1][i + 1] != WelltrisInterface.EMPTY)) {
			      tempBase[j][i] &= ~PolyominoImage.RIGHT_DOWN;
			    }
			  }
			}
		/* Now draw */
		for (i = 0; i < curThing.size; i++)
			for (j = 0; j < curThing.size; j++)
				if (tris[curThing.unitsIndex].poly
				      [curThing.polyNumber].shape[j][i] != 0) {
				  xi = (curThing.position.x + i) % perimeter;
				  yj = curThing.position.y + j;
				  if (yj >= depth) {
				    Point baseCoord = wallToBase(xi, yj);
				    if (tempBase[baseCoord.y][baseCoord.x] >= 0) {
				      /* drawUnit(g, polyominoImage.polyPixmap(baseCoord.x, baseCoord.y,
					curThing, tempBase),
					curThing.colorNumber,
					baseCoord.x, baseCoord.y); */
				      /* This uses precomputed values */
				      drawUnit(g,
					tempBase[baseCoord.y][baseCoord.x],
					curThing.colorNumber,
					baseCoord.x, baseCoord.y);
				      /* negate for erase */
				      tempBase[baseCoord.y][baseCoord.x] =
					-tempBase[baseCoord.y][baseCoord.x] - 2;
				    }
				  } else if (yj >= 0) {
					int rotPmid = rotateIndex(
						PolyominoImage.polyPixmap(i, j,
						tris[curThing.unitsIndex].poly[curThing.polyNumber].shape,
						curThing.size),
						xi, yj);

				    drawTrapazoid(g,
				      checkWallSides(checkWallBottom(rotPmid, yj), xi),
				      PolyominoImage.colorThings[curThing.colorNumber],
				      ((grid) ? Color.black : background),
				      xi, yj, PolyominoImage.DRAW_COLOR);
				  }
				}
		/* Erase */
		for (i = 0; i < oldThing.size; i++)
			for (j = 0; j < oldThing.size; j++) {
				xi = (oldThing.position.x + i) % perimeter;
				yj = oldThing.position.y + j;
				ox = (xi - curThing.position.x + perimeter) % perimeter;
				oy = yj - curThing.position.y;
				if (yj >= depth) {
				  Point baseCoord = wallToBase(xi, yj);
				  if (tris[oldThing.unitsIndex].poly
				      [oldThing.polyNumber].shape[j][i] != 0 &&
				      ((ox >= curThing.size) || (oy < 0) ||
				      (oy >= curThing.size) ||
				      tempBase[baseCoord.y][baseCoord.x] == WelltrisInterface.EMPTY)) {
				    clearUnit(g, background, baseCoord.x, baseCoord.y);
				  }
				} else if (yj >= 0 &&
				      (tris[oldThing.unitsIndex].poly
				      [oldThing.polyNumber].shape[j][i] != 0) &&
				      ((ox >= curThing.size) || (oy < 0) ||
				      (oy >= curThing.size) || tris[curThing.unitsIndex].poly
				      [curThing.polyNumber].shape[oy][ox] == 0)) {
				    drawTrapazoid(g,
				      PolyominoImage.NUM_BITMAPS - 1,
				      background,
				      ((grid) ? Color.black : background),
				      xi, yj, PolyominoImage.CLEAR_COLOR);
				}
			}
	}

	void drawUnit(int pmid, int cid, int i, int j) {
		Graphics g = this.getGraphics();

		if (j < 0) {
			return;
		}
		if (j < depth) {
			drawTrapazoid(g, pmid,
				PolyominoImage.colorThings[cid],
				((grid) ? Color.black : background),
				i, j, PolyominoImage.DRAW_COLOR);
		} else {
			Point baseCoord = wallToBase(i, j);
			drawUnit(g, pmid, cid, baseCoord.x, baseCoord.y);
		}
	}

	void drawNext(JFrame app) {
		((WelltrisFrame) app).drawNext(nextThing, tris);
	}

	void clearNext(JFrame app) {
		((WelltrisFrame) app).clearNext(nextThing, tris);
	}

	void redoNext() {
		int diagonal = (nextSpec.diagonal) ? 1 : 0;
		int turnStyle =
			tris[nextSpec.units - PolyominoThing.MIN_UNITS].mode
				[diagonal].turnStyle;
		int nextStart, i, diffPolys;

		nextThing.unitsIndex = nextSpec.units - PolyominoThing.MIN_UNITS +
			((nextThing.getBonus()) ? 1 : 0);
		diffPolys = tris[nextThing.unitsIndex].mode[diagonal].number
			[(nextThing.getBonus()) ? PolyominoThing.NOMIX :
			((nextSpec.mixed) ? 1 : 0)][turnStyle];
		if (diffPolys == 0) {
			System.out.println("no poly units=" + nextThing.unitsIndex + PolyominoThing.MIN_UNITS
				+ ", diagonal=" + nextSpec.diagonal
				+ ", mixed=" + nextSpec.mixed
				+ ", bonus=" + nextThing.getBonus());
			return;
		}
		nextStart = nextThing.getRandomNumber() % diffPolys;
		nextThing.colorNumber = nextStart;
		if (nextSpec.mixed && !nextThing.getBonus()) {
			nextThing.unitsIndex = 0;
			while (nextStart >= tris[nextThing.unitsIndex].mode
				  [diagonal].number[PolyominoThing.NOMIX][turnStyle]) {
				nextStart -= tris[nextThing.unitsIndex].mode
				  [diagonal].number[PolyominoThing.NOMIX][turnStyle];
				nextThing.unitsIndex++;
			}
		}
		nextThing.polyNumber = tris[nextThing.unitsIndex].mode
			[diagonal].start[nextStart][turnStyle];
		if (tris[nextThing.unitsIndex].mode[diagonal].turnStyle >
				PolyominoThing.NONE)
			for (i = 0; i < nextThing.getRotation(); i++)
				nextThing.polyNumber =
					tris[nextThing.unitsIndex].poly
					[nextThing.polyNumber].rotation;
		if (tris[nextThing.unitsIndex].mode[diagonal].turnStyle ==
				PolyominoThing.ALL)
			for (i = 0; i < nextThing.getReflection(); i++)
				nextThing.polyNumber =
					tris[nextThing.unitsIndex].poly
					[nextThing.polyNumber].reflection;
		nextThing.size = tris[nextThing.unitsIndex].poly
			[nextThing.polyNumber].size;
		nextThing.position.y = -tris[nextThing.unitsIndex].poly
			[nextThing.polyNumber].startHeight[(suddenAppear) ? 1 : 0];
		if (debug)
			System.out.println("Size: " + nextThing.size +
				", Poly: " + nextThing.polyNumber +
				", Color: " + nextThing.colorNumber);
	}

	boolean getPlainPiece() {
		return polyominoImage.getPlainPiece();
	}

	void setPlainPiece(boolean plainPiece) {
		polyominoImage.setPlainPiece(plainPiece);
		redrawPaint = true;
	}

	void banner(String msg) {
		((WelltrisFrame) frame).displayMessage(msg);
	}

	void putUnit() {
		int i, j, xi, yj;
		int[][] tempBase = new int[diameter][diameter];

		/* Initialize */
		for (i = 0; i < diameter; i++)
			for (j = 0; j < diameter; j++)
				tempBase[j][i] = WelltrisInterface.EMPTY;
		/* Set pieces and overlap them if necessary */
		for (i = 0; i < curThing.size; i++)
			for (j = 0; j < curThing.size; j++)
				if (tris[curThing.unitsIndex].poly
				      [curThing.polyNumber].shape[j][i] != 0) {
				  xi = (curThing.position.x + i) % perimeter;
				  yj = curThing.position.y + j;
				  if (yj >= depth) {
				    int rotPmid = rotateIndex(PolyominoImage.polyPixmap(i, j,
				      tris[curThing.unitsIndex].poly[curThing.polyNumber].shape,
				      curThing.size), xi, yj);

				    Point baseCoord = wallToBase(xi, yj);
				    xi = checkWall(rotPmid, baseCoord.x, baseCoord.y);
				    tempBase[baseCoord.y][baseCoord.x] &= xi;
				  } else if (yj >= 0) {
				    int rotPmid = rotateIndex(PolyominoImage.polyPixmap(i, j,
				      tris[curThing.unitsIndex].poly[curThing.polyNumber].shape,
				      curThing.size), xi, yj);

				    wall[yj][xi].pmid =
				      checkWallSides(checkWallBottom(rotPmid, yj), xi);
				    wall[yj][xi].cid = curThing.colorNumber;
				    frozenWall[xi / diameter] = WelltrisInterface.MAX_SIDES;
				  }
				}
		/* Test to make corner split neighbor awareness */
		for (i = 0; i < diameter; i++)
			for (j = 0; j < diameter; j++) {
			  if (tempBase[j][i] >= 0) {
			    if (j > 0 && (tempBase[j - 1][i] != WelltrisInterface.EMPTY)) {
				tempBase[j][i] &= ~PolyominoImage.UP;
			    }
			    if ((j < diameter - 1) &&
					(tempBase[j + 1][i] != WelltrisInterface.EMPTY)) {
				tempBase[j][i] &= ~PolyominoImage.DOWN;
			    }
			    if (i > 0 && (tempBase[j][i - 1] != WelltrisInterface.EMPTY)) {
				tempBase[j][i] &= ~PolyominoImage.LEFT;
			    }
			    if ((i < diameter - 1) &&
					(tempBase[j][i + 1] != WelltrisInterface.EMPTY)) {
				tempBase[j][i] &= ~PolyominoImage.RIGHT;
			    }
			    if (i > 0 && j > 0 &&
					(tempBase[j - 1][i - 1] != WelltrisInterface.EMPTY)) {
				tempBase[j][i] &= ~PolyominoImage.LEFT_UP;
			    }
			    if (i > 0 && (j < diameter - 1) &&
					(tempBase[j + 1][i - 1] != WelltrisInterface.EMPTY)) {
				tempBase[j][i] &= ~PolyominoImage.LEFT_DOWN;
			    }
			    if ((i < diameter - 1) && j > 0 &&
					(tempBase[j - 1][i + 1] != WelltrisInterface.EMPTY)) {
				tempBase[j][i] &= ~PolyominoImage.RIGHT_UP;
			    }
			    if ((i < diameter - 1) && (j < diameter - 1) &&
					(tempBase[j + 1][i + 1] != WelltrisInterface.EMPTY)) {
				tempBase[j][i] &= ~PolyominoImage.RIGHT_DOWN;
			    }
			  }
			}
		for (i = 0; i < diameter; i++)
			for (j = 0; j < diameter; j++)
				if (tempBase[j][i] != WelltrisInterface.EMPTY) {
					base[j][i].pmid = tempBase[j][i];
					base[j][i].cid = curThing.colorNumber;
				}
	}

	boolean overlapping() {
		int i, j, xi, yj;
		int x = curThing.position.x, y = curThing.position.y;

		if (wall == null)
			return false;
		for (i = 0; i < curThing.size; i++)
			for (j = 0; j < curThing.size; j++)
			  if (tris[curThing.unitsIndex].poly
			      [curThing.polyNumber].shape[j][i] != 0) {
			    xi = (x + i) % perimeter;
			    yj = y + j;
			    if (yj < depth) {
			      if (frozenWall[xi / diameter] != 0)
				return true;
			      if (suddenAppear) { /* creationalism */
/* This method does not allow turning polyomino to an area above screen */
				if ((yj < 0) || (wall[yj][xi].pmid >= 0))
				  return true;
			      } else { /* evolution */
/* This method one can turn polyomino to an area above of screen, also
   part of the polyomino may not be visible initially */
				if ((yj >= 0) && (wall[yj][xi].pmid >= 0))
				  return true;
			      }
			    } else if (yj < depth + diameter) {
			      Point baseCoord = wallToBase(xi, yj);
			      if (base[baseCoord.y][baseCoord.x].pmid >= 0)
				return true;
			    } else
			      return true;
			  }
		return false;
	}

	boolean atBottom() {
		int i, j, xi, yj;
		int x = curThing.position.x, y = curThing.position.y;

		for (i = 0; i < curThing.size; i++)
			for (j = 0; j < curThing.size; j++)
			  if (tris[curThing.unitsIndex].poly
			      [curThing.polyNumber].shape[j][i] != 0) {
			    xi = (x + i) % perimeter;
			    yj = y + j;
			    if (yj < -1)
			      return false;
			    if (yj < depth - 1) {
			      if (frozenWall[xi / diameter] != 0)
				return true;
			      if (wall[yj + 1][xi].pmid >= 0)
				return true;
			    } else if (yj < depth + diameter - 1) {
			      Point baseCoord = wallToBase(xi, yj + 1);
			      if (base[baseCoord.y][baseCoord.x].pmid >= 0)
				return true;
			    } else
			      return true;
			  }
		return false;
	}

	boolean atBaseFully() {
		for (int i = 0; i < curThing.size; i++)
			for (int j = 0; j < curThing.size; j++)
			  if (curThing.position.y + j < depth && tris[curThing.unitsIndex].poly
			      [curThing.polyNumber].shape[j][i] != 0)
			     return false;
		return true;
	}

	boolean atBasePartially() {
		for (int i = 0; i < curThing.size; i++)
			for (int j = 0; j < curThing.size; j++)
			  if (curThing.position.y + j >= depth && tris[curThing.unitsIndex].poly
			      [curThing.polyNumber].shape[j][i] != 0)
			     return true;
		return false;
	}

	boolean wallChange(Thing oldThing, Thing newThing) {
		int	     w = -1, i, j;
		int	     x = curThing.position.x;

		for (i = 0; i < oldThing.size; i++)
			for (j = 0; j < oldThing.size; j++)
			  if (tris[oldThing.unitsIndex].poly[oldThing.polyNumber].shape[j][i] != 0) {
			    if (w == -1)
			      w = ((x + i) % perimeter) / diameter;
			    else if (w != ((x + i) % perimeter) / diameter)
			      return true;
			  }
		for (i = 0; i < newThing.size; i++)
			for (j = 0; j < newThing.size; j++)
			  if (tris[newThing.unitsIndex].poly[newThing.polyNumber].shape[j][i] != 0 &&
			      (w != ((x + i) % perimeter) / diameter))
			    return true;
		return false;
	}

	void tryMove(int move) {
		Thing oldThing = new Thing();
		int i;
		boolean skip = false;

		oldThing.clone(curThing);
		switch (move) {
		case WelltrisInterface.FALL:
			curThing.position.y++;
			break;
		case WelltrisInterface.DROP:
			do {
				curThing.position.y++;
				if (!dropPaused)
					score += level + fill + ((grid) ? 0 : 1);
			} while (!overlapping());
			curThing.position.y--;
			break;

		case WelltrisInterface.MOVE_LEFT:
			curThing.position.x = (curThing.position.x + perimeter - 1) % perimeter;
			skip = (atBaseFully() ||
				(atBasePartially() && wallChange(oldThing, curThing)));
			break;

		case WelltrisInterface.MOVE_RIGHT:
			curThing.position.x = (curThing.position.x + 1) % perimeter;
			skip = (atBaseFully() ||
				(atBasePartially() && wallChange(oldThing, curThing)));
			break;

		case WelltrisInterface.ROTATE:
			if (tris[curSpec.units - PolyominoThing.MIN_UNITS].mode
				[(curSpec.diagonal) ? 1 : 0].turnStyle > PolyominoThing.NONE) {
			  if (cw)
			    for (i = 0; i < WelltrisInterface.MAX_SIDES - 1; i++)
			      curThing.polyNumber = tris[curThing.unitsIndex].poly
				[curThing.polyNumber].rotation;
			  else /* ccw */
			    curThing.polyNumber = tris[curThing.unitsIndex].poly
			      [curThing.polyNumber].rotation;
			  curThing.position.x = oldThing.position.x;
			  curThing.position.y = oldThing.position.y;
			  skip = (atBasePartially() && wallChange(oldThing, curThing));
			}
			break;
		case WelltrisInterface.REFLECT: /* reflect on y axis */
			if (tris[curSpec.units - PolyominoThing.MIN_UNITS].mode
			    [(curSpec.diagonal) ? 1 : 0].turnStyle == PolyominoThing.ALL) {
			  curThing.polyNumber = tris[curThing.unitsIndex].poly
			      [curThing.polyNumber].reflection;
			  curThing.position.x = oldThing.position.x;
			  curThing.position.y = oldThing.position.y;
			  skip = (atBasePartially() && wallChange(oldThing, curThing));
			}
			break;
		default:
		}
		if (redrawPaint) {
			drawField();
			redrawPaint = false;
		}
		if (framePaint) {
			drawFrame(focus);
			framePaint = false;
		}
		if (skip || overlapping())
			curThing = oldThing;
		else
			drawThingDiff(oldThing);
	}

	void initFill() {
		int i, j, col, fixedUnits;
		int turnStyle = tris[curSpec.units - PolyominoThing.MIN_UNITS].mode
			[(curSpec.diagonal) ? 1 : 0].turnStyle;
		int mixed = (nextSpec.mixed && WelltrisInterface.mixedLegal(nextSpec.units, nextSpec.mixed)) ? 1 : 0;
		int numberPolyominoes = tris[nextSpec.units - PolyominoThing.MIN_UNITS].mode
			[(nextSpec.diagonal) ? 1 : 0].number[mixed][turnStyle];

		if (fixedFill) {
			/* This algorithm randomly picks fixedUnits per row
			 * (4 for DEFAULTDIAMETER, ie. 8/2)
			 * This is more in keeping with the original */
			fixedUnits = diameter / 2;
			for (j = 0; j < depth; j++) {
				for (i = 0; i < perimeter; i++) {
					wall[j][i].pmid = -1;
					wall[j][i].cid = 0;
				}
				if (j >= depth - fill)
					for (i = 0; i < fixedUnits; i++) {
						do
							col = generator.nextInt(perimeter);
						while (wall[j][col].pmid != -1);
						wall[j][col].pmid = PolyominoImage.NUM_BITMAPS - 1;
						wall[j][col].cid = generator.nextInt(numberPolyominoes);
					}
			}
		} else {
			/* This algorithm randomly fills boxes to a certain level
			   The problem with this is that it could potentially fill
			   up a whole row */
			for (j = 0; j < depth; j++)
				for (i = 0; i < perimeter; i++)
					if ((j >= depth - fill) &&
							(generator.nextInt(8) == 0)) {
						wall[j][i].pmid = PolyominoImage.NUM_BITMAPS - 1;
						wall[j][i].cid = generator.nextInt(numberPolyominoes);
					} else {
						wall[j][i].pmid = -1;
						wall[j][i].cid = 0;
					}
		}
	}

	void fillLines() {
		int turnStyle = tris[curSpec.units - PolyominoThing.MIN_UNITS].mode
			[(curSpec.diagonal) ? 1 : 0].turnStyle;
		int dir = generator.nextInt(2);  // spin direction
		int pmid = PolyominoImage.NUM_BITMAPS - 1;

		for (int k = 0; k < depth + diameter; k++) {
			for (int i = 0; i < perimeter; i++) {
				int j = i;

				if (dir == 1)
				  j = perimeter - 1 - i;
				drawUnit(pmid,
				  generator.nextInt(tris[curSpec.units - PolyominoThing.MIN_UNITS].mode
				  [(curSpec.diagonal) ? 1 : 0].number[(curSpec.mixed) ? 1 : 0]
				  [turnStyle]), j, k);
			}
		}
	}

	void freezeWall(int w) {
		Graphics g = this.getGraphics();

		for (int j = 0; j < depth; j++)
			for (int i = 0; i < diameter; i++) {
				Color color;
				if (wall[j][w * diameter + i].pmid < 0)
					color = background;
				else
					color = PolyominoImage.colorThings[wall[j][w * diameter + i].cid];
				if (frozenWall[w] == 0) {
					drawTrapazoid(g,
						wall[j][w * diameter + i].pmid,
						color,
						((grid) ? Color.black : background),
						w * diameter + i, j,
						((wall[j][w * diameter + i].pmid < 0) ?
						PolyominoImage.CLEAR_COLOR : PolyominoImage.DRAW_COLOR));
				} else { // Simulate Xor
					color = new Color(255 - color.getRed(),
						255 - color.getGreen(), 255 - color.getBlue());
					drawTrapazoid(g,
						wall[j][w * diameter + i].pmid,
						color,
						((grid) ? Color.black : background),
						w * diameter + i, j,
						((wall[j][w * diameter + i].pmid < 0) ?
						PolyominoImage.XOR_CLEAR_COLOR : PolyominoImage.XOR_COLOR));
				}
			}
		if (((WelltrisFrame) frame).getToggleSound())
			((WelltrisFrame) frame).playDripAudio();
	}

	static void unfreezeWalls() {
		int i;

		for (i = 0; i < WelltrisInterface.MAX_SIDES; i++)
			frozenWall[i] = 0;
	}

	void dropWall(int w) {
		Graphics g = this.getGraphics();
		int i, j, k, l, lines = 0;
		boolean fits;

		for (j = depth - 1; j >= 0 && lines == 0; j--)
			for (i = 0; i < diameter; i++)
				if (wall[j][w * diameter + i].pmid != WelltrisInterface.EMPTY)
					lines = depth - j;
		if (lines > 0) {
			fits = true;
			j = 0;
			while (j < (diameter / 2 + lines - 1) && fits) {
				j++;
				for (l = depth - j; l < depth; l++)
				  for (i = 0; i < diameter; i++)
				    if (wall[l][w * diameter + i].pmid != WelltrisInterface.EMPTY) {
				      Point baseCoord = wallToBase(w * diameter + i, l + j);
				      if (base[baseCoord.y][baseCoord.x].pmid != WelltrisInterface.EMPTY)
					fits = false;
				    }
			}
			if (!fits)
				j--;
			if (j > 0) {
				for (l = depth - 1; l >= 0; l--)
				  for (i = 0; i < diameter; i++)
				    if (wall[l][w * diameter + i].pmid != WelltrisInterface.EMPTY) {
				      k = w * diameter + i;
				      if (l + j >= depth) {
					Point baseCoord = wallToBase(k, l + j);
					base[baseCoord.y][baseCoord.x].cid = wall[l][k].cid;
					base[baseCoord.y][baseCoord.x].pmid = checkWall(rotateIndex(wall[l][k].pmid,
					  k, l + j), baseCoord.x, baseCoord.y);
				       wall[l][k].pmid = WelltrisInterface.EMPTY;
				       drawTrapazoid(g,
					  PolyominoImage.NUM_BITMAPS - 1,
					  background,
					 ((grid) ? Color.black : background),
					 k, l, PolyominoImage.CLEAR_COLOR);
				       drawUnit(g, base[baseCoord.y][baseCoord.x].pmid, base[baseCoord.y][baseCoord.x].cid,
					 baseCoord.x, baseCoord.y);
				     } else {
				       wall[l + j][k].cid = wall[l][k].cid;
				       wall[l + j][k].pmid = checkWallBottom(wall[l][k].pmid, l + j);
				      wall[l][k].pmid = WelltrisInterface.EMPTY;
				      drawTrapazoid(g,
					PolyominoImage.NUM_BITMAPS - 1,
					background,
					((grid) ? Color.black : background),
					k, l, PolyominoImage.CLEAR_COLOR);
				      drawTrapazoid(g, wall[l + j][k].pmid,
					PolyominoImage.colorThings[wall[l][k].cid],
					((grid) ? Color.black : background),
					k, l + j, PolyominoImage.DRAW_COLOR);
				    }
				  }
			}
		}
		if (((WelltrisFrame) frame).getToggleSound())
			((WelltrisFrame) frame).playDripAudio();
	}

	boolean allFrozen() {
		int w;

		for (w = 0; w < WelltrisInterface.MAX_SIDES; w++)
			if (frozenWall[w] == WelltrisInterface.MAX_SIDES) {
				freezeWall(w);
			}
		for (w = 0; w < WelltrisInterface.MAX_SIDES; w++)
			if ((frozenWall[w] == 0))
				return false;
		return true;
	}

	void checkFreeze() {
		int w;

		if (generator.nextInt(2) == 1) {
			for (w = 0; w < WelltrisInterface.MAX_SIDES; w++) {
				if ((frozenWall[w] == 0))
					dropWall(w);
				else {
					frozenWall[w]--;
					if ((frozenWall[w] == 0)) {
						freezeWall(w);
						dropWall(w);
					}
				}
			}
		} else {
			for (w = WelltrisInterface.MAX_SIDES - 1; w >= 0; w--) {
				if ((frozenWall[w] == 0))
					dropWall(w);
				else {
					frozenWall[w]--;
					if ((frozenWall[w] == 0)) {
						freezeWall(w);
						dropWall(w);
					}
				}
			}
		}
	}

	int checkLines() {
		Graphics g = this.getGraphics();
		Field[][] tempBase = new Field[diameter][diameter];
		int[][] lSet = new int[2][diameter];
		int[] nSet = new int[2];
		int i, j, k;

		nSet[0] = nSet[1] = 0;

		for (j = 0; j < diameter; j++)
			for (i = 0; i < diameter; i++) {
				tempBase[j][i] = new Field();
				tempBase[j][i].pmid = base[j][i].pmid;
				tempBase[j][i].cid = base[j][i].cid;
			}
		for (j = 0; j < diameter; j++) {
			/* Find filled rows */
			lSet[0][j] = 0;
			for (i = 0; i < diameter; i++)
				if (tempBase[j][i].pmid >= 0)
					lSet[0][j]++;
			if (lSet[0][j] == diameter)
				nSet[0]++;
			/* Find filled columns */
			lSet[1][j] = 0;
			for (i = 0; i < diameter; i++)
				if (tempBase[i][j].pmid >= 0)
					lSet[1][j]++;
			if (lSet[1][j] == diameter)
				nSet[1]++;
		}

		if ((nSet[0] != 0) || (nSet[1] != 0)) {

			for (k = 0; k < 2 * WelltrisInterface.NUM_FLASHES; k++) {
				for (j = 0; j < diameter; j++) {
					if (lSet[0][j] == diameter)
					  for (i = 0; i < diameter; i++) {
					    Color color = PolyominoImage.colorThings[tempBase[j][i].cid];
					    if (k % 2 == 0) // Simulate Xor
					      color = new Color(255 - color.getRed(),
						255 - color.getGreen(), 255 - color.getBlue());
					    if (tempBase[j][i].pmid < 0)
					      System.err.println("checkLines1 tempBase " + tempBase[j][i].pmid +
						", i "  + i + ", j " + j +
						", lSet " + lSet[0][j] +
						", nSet " + nSet[0]);
					    else
					      drawUnit(g, tempBase[j][i].pmid, color, i, j);
					  }
					if (lSet[1][j] == diameter)
					  for (i = 0; i < diameter; i++) {
					    Color color = PolyominoImage.colorThings[tempBase[i][j].cid];
					    if (k % 2 == 0) // Simulate Xor
					      color = new Color(255 - color.getRed(),
						255 - color.getGreen(), 255 - color.getBlue());
					    if (tempBase[i][j].pmid < 0)
					      System.err.println("checkLines2 tempBase " + tempBase[i][j].pmid +
						", j "  + j + ", i " + i +
						", lSet " + lSet[1][j] +
						", nSet " + nSet[1]);
					    else
					      drawUnit(g, tempBase[i][j].pmid, color, j, i);
					  }
				}
				try {
					Thread.sleep(100L);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}

			if (nSet[0] != 0) {
				for (j = 0; j < diameter / 2; j++)
					if (lSet[0][j] == diameter)
					  for (i = 0; i < diameter; i++) {
					    for (k = j; k > 0; k--) {
					      tempBase[k][i].pmid =
						tempBase[k - 1][i].pmid;
					      tempBase[k][i].cid =
						tempBase[k - 1][i].cid;
					    }
					    tempBase[0][i].pmid = WelltrisInterface.EMPTY;
					    if ((j > 0) &&
						(tempBase[j][i].pmid >= 0)) {
					      tempBase[j][i].pmid =
						PolyominoImage.CHECKDOWN(
						  tempBase[j][i].pmid);
					    }
					    if (tempBase[j + 1][i].pmid >= 0)
					      tempBase[j + 1][i].pmid =
						PolyominoImage.CHECKUP(
						  tempBase[j + 1][i].pmid);
					  }
				for (j = diameter - 1; j >= diameter / 2; j--)
					if (lSet[0][j] == diameter)
					  for (i = 0; i < diameter; i++) {
					    for (k = j; k < diameter - 1; k++) {
					      tempBase[k][i].pmid =
						tempBase[k + 1][i].pmid;
					      tempBase[k][i].cid =
						tempBase[k + 1][i].cid;
					    }
					    tempBase[diameter - 1][i].pmid = WelltrisInterface.EMPTY;
					    if ((j < diameter - 1) &&
						(tempBase[j][i].pmid >= 0)) {
					      tempBase[j][i].pmid =
						PolyominoImage.CHECKUP(
						 tempBase[j][i].pmid);
					    }
					    if (tempBase[j - 1][i].pmid >= 0)
					      tempBase[j - 1][i].pmid =
						PolyominoImage.CHECKDOWN(
						  tempBase[j - 1][i].pmid);
					  }
			}
			if (nSet[1] != 0) {
				for (i = 0; i < diameter / 2; i++)
					if (lSet[1][i] == diameter)
					  for (j = 0; j < diameter; j++) {
					    for (k = i; k > 0; k--) {
					      tempBase[j][k].pmid =
						tempBase[j][k - 1].pmid;
					      tempBase[j][k].cid =
						tempBase[j][k - 1].cid;
					    }
					    tempBase[j][0].pmid = WelltrisInterface.EMPTY;
					    if ((i > 0) &&
						(tempBase[j][i].pmid >= 0)) {
					      tempBase[j][i].pmid =
						PolyominoImage.CHECKRIGHT(
						  tempBase[j][i].pmid);
					    }
					    if (tempBase[j][i + 1].pmid >= 0)
					      tempBase[j][i + 1].pmid =
						PolyominoImage.CHECKLEFT(
						  tempBase[j][i + 1].pmid);
					  }
				for (i = diameter - 1; i >= diameter / 2; i--)
					if (lSet[1][i] == diameter)
					  for (j = 0; j < diameter; j++) {
					    for (k = i; k < diameter - 1; k++) {
					      tempBase[j][k].pmid =
						tempBase[j][k + 1].pmid;
					      tempBase[j][k].cid =
						tempBase[j][k + 1].cid;
					    }
					    tempBase[j][diameter - 1].pmid =
					      WelltrisInterface.EMPTY;
					    if ((i < diameter - 1) &&
						(tempBase[j][i].pmid >= 0)) {
					      tempBase[j][i].pmid =
						PolyominoImage.CHECKLEFT(
						  tempBase[j][i].pmid);
					    }
					    if (tempBase[j][i - 1].pmid >= 0)
					      tempBase[j][i - 1].pmid =
						PolyominoImage.CHECKRIGHT(
						  tempBase[j][i - 1].pmid);
					  }
			}
			for (j = 0; j < diameter; j++)
				for (i = 0; i < diameter; i++)
					if ((tempBase[j][i].cid != base[j][i].cid) ||
					    (tempBase[j][i].pmid != base[j][i].pmid)) {
					  base[j][i].pmid = tempBase[j][i].pmid;
					  base[j][i].cid = tempBase[j][i].cid;
					  if (base[j][i].pmid >= 0)
					    drawUnit(g, base[j][i].pmid, base[j][i].cid, i, j);
					  else
					    clearUnit(g, background, i, j);
					}
		}
		if (((WelltrisFrame) frame).getToggleSound())
			((WelltrisFrame) frame).playBumpAudio();
		return (nSet[0] + nSet[1]);
	}

	void drawGrid() {
		Graphics g = this.getGraphics();
		Color color;

		if (g == null)
			return;
		if (grid)
			color = Color.black;
		else
			color = getBackground();
		g.setColor(color);
		for (int i = 0; i < perimeter; i++)
			for (int k = 0; k < depth; k++)
				drawTrapazoidBorder(g, i, k);
		/* int i, k, square_x, sum, sum_var;
		sum = sum_var = diameter * boxSize + delta - 1;
		square_x = 0;
		for (i = 0; i < diameter; i++) {
			g.drawLine(square_x, 0, sum_var, sum);
			g.drawLine(endPt - 1, square_x, endPt - 1 - sum, sum_var);
			g.drawLine(endPt - 1 - square_x, endPt - 1,
				endPt - 1 - sum_var, endPt - 1 - sum);
			g.drawLine(0, endPt - 1 - square_x, sum, endPt - 1 - sum_var);
			square_x += 3 * boxSize;
			sum_var += boxSize;
		}
		sum = endPt;
		square_x = 0;
		k = (diameter * boxSize) / depth;
		for (i = 0; i <= depth; i++) {
			g.fillRect(square_x, square_x, 1, sum);
			g.fillRect(square_x, square_x, sum, 1);
			g.fillRect(square_x + sum - 1, square_x, 1, sum);
			g.fillRect(square_x, square_x + sum - 1, sum, 1);
			square_x += k;
			sum -= (2 * k);
		}*/
		/* sum = diameter * boxSize + delta;
		square_x = square_y = sum;
		for (i = 0; i <= diameter ; i++) {
			g.fillRect(square_x, square_y, 1, sum);
			square_x += boxSize;
		}
		square_x = square_y = sum;
		for (i = 0; i <= diameter ; i++) {
			g.fillRect(square_x, square_y, sum, 1);
			square_y += boxSize;
		} */
	}

	void drawTrapazoid(Graphics g, int pmid, Color cid, Color bid,
			int i, int j, int draw) {
		int ulx, uly, lrx, lry, trapBase, plateau, trapBaseWidth, trapEndPt, sideWall;
		int newI = i;

		trapBaseWidth =  diameter * boxSize + 1;
		trapEndPt = 3 * trapBaseWidth;
		sideWall = newI / diameter;
		newI -= (sideWall * diameter);
		ulx = (trapEndPt +  (2 * newI - diameter) * ((depth / 2 +
			2 * boxSize * (depth - j)) / depth + boxSize)) / 2;
		uly = (j * trapBaseWidth + depth / 2) / depth;
		lrx = (trapEndPt + (2  + 2 * newI  - diameter) * ((depth / 2 +
			2 * boxSize * (depth - j - 1)) / depth + boxSize)) / 2;
		lry = ((j + 1) * trapBaseWidth + depth / 2) / depth;
		trapBase = boxSize + ((depth - j) * 2 * boxSize + depth / 2) /
			depth;
		plateau = boxSize + ((depth - j - 1) * 2 * boxSize +
			depth / 2) / depth;
		switch (sideWall) {
		case 0:
			polyominoImage.drawWallUnit(g, pmid, cid, bid,
				fieldOffset.x, fieldOffset.y,
				ulx + 1, uly, trapBase, 0,
				lrx + 1, lry, plateau, 0,
				draw);
			break;
		case 1:
			polyominoImage.drawWallUnit(g, pmid, cid, bid,
				fieldOffset.x, fieldOffset.y,
				trapEndPt - uly, ulx + 1, 0, trapBase,
				trapEndPt - lry, lrx + 1, 0, plateau,
				draw);
			break;
		case 2:
			polyominoImage.drawWallUnit(g,
				pmid + PolyominoImage.NUM_BITMAPS,
				cid, bid,
				fieldOffset.x, fieldOffset.y,
				trapEndPt - 1 - ulx, trapEndPt - uly, -trapBase, 0,
				trapEndPt - 1 - lrx, trapEndPt - lry, -plateau, 0,
				draw);
			break;
		case 3:
			polyominoImage.drawWallUnit(g,
				pmid + PolyominoImage.NUM_BITMAPS,
				cid, bid,
				fieldOffset.x, fieldOffset.y,
				uly, trapEndPt - 1 - ulx, 0, -trapBase,
				lry, trapEndPt - 1 - lrx, 0, -plateau,
				draw);
			break;
		default:
			System.err.println("trapazoid kinds range 0-3, got " +
				sideWall + ".");
		}
	}

	void drawTrapazoidBorder(Graphics g, int i, int j) {
		int ulx, uly, lrx, lry, trapBase, plateau, trapBaseWidth, trapEndPt, sideWall;
		int newI = i;

		trapBaseWidth =  diameter * boxSize + 1;
		trapEndPt = 3 * trapBaseWidth;
		sideWall = newI / diameter;
		newI -= (sideWall * diameter);
		ulx = (trapEndPt +  (2 * newI - diameter) * ((depth / 2 +
			2 * boxSize * (depth - j)) / depth + boxSize)) / 2;
		uly = (j * trapBaseWidth + depth / 2) / depth;
		lrx = (trapEndPt + (2  + 2 * newI  - diameter) * ((depth / 2 +
			2 * boxSize * (depth - j - 1)) / depth + boxSize)) / 2;
		lry = ((j + 1) * trapBaseWidth + depth / 2) / depth;
		trapBase = boxSize + ((depth - j) * 2 * boxSize + depth / 2) /
			depth;
		plateau = boxSize + ((depth - j - 1) * 2 * boxSize +
			depth / 2) / depth;
		switch (sideWall) {
		case 0:
			PolyominoImage.drawWallUnitBorder(g,
				// pmid,
				fieldOffset.x, fieldOffset.y,
				ulx + 1, uly, trapBase, 0,
				lrx + 1, lry, plateau, 0);
			break;
		case 1:
			PolyominoImage.drawWallUnitBorder(g,
				// pmid,
				fieldOffset.x, fieldOffset.y,
				trapEndPt - uly, ulx + 1, 0, trapBase,
				trapEndPt - lry, lrx + 1, 0, plateau);
			break;
		case 2:
			PolyominoImage.drawWallUnitBorder(g,
				// pmid,
				fieldOffset.x, fieldOffset.y,
				trapEndPt - 1 - ulx, trapEndPt - uly, -trapBase, 0,
				trapEndPt - 1 - lrx, trapEndPt - lry, -plateau, 0);
			break;
		case 3:
			PolyominoImage.drawWallUnitBorder(g,
				// pmid,
				fieldOffset.x, fieldOffset.y,
				uly, trapEndPt - 1 - ulx, 0, -trapBase,
				lry, trapEndPt - 1 - lrx, 0, -plateau);
			break;
		default:
			System.err.println("trapazoid kinds range 0-3, got " +
				sideWall + ".");
		}
	}

	void clearUnit(Graphics g, Color color, int i, int j) {
		g.setColor(color);
		g.fillRect(i * boxSize + baseWidth + 1 + fieldOffset.x,
			j * boxSize + baseWidth + 1 + fieldOffset.y,
			boxSize, boxSize);
	}

// ELIMINATE ME later?
	void drawUnit(Graphics g, int pmid, Color color, int i, int j) {
		polyominoImage.drawUnit(g, false, pmid, color,
			baseWidth + 1 + fieldOffset.x,
			baseWidth + 1 + fieldOffset.y,
			i, j, boxSize, boxSize);
	}

	void drawUnit(Graphics g, int pmid, int cid, int i, int j) {
		drawUnit(g, pmid, PolyominoImage.colorThings[cid], i, j);
	}

	@Override
	public void update(Graphics g) {
		paint(g); /* no erase */
	}

	@Override
	public void paint(Graphics g) {
		if (firstPaint) {
			firstPaint = false;
			resizePaint = true;
		}
		if (resizePaint) {
			clearField();
			resizePaint = false;
		}
		if (framePaint) {
			drawFrame(g, focus);
			framePaint = false;
		}
		redrawPaint = false;
		drawField();
		drawThing();
	}

	void paintNow() {
		Graphics g = getGraphics();

		try {
			paint(g);
		} catch (Exception e) {
			e.printStackTrace();
		} finally {
			g.dispose();
		}
	}

	transient PolyominoThing polyominoThing = null;
	transient PolyominoThingFallback polyominoThingFallback = null;
	final void readPolyominoes() {
		polyominoThing = new PolyominoThing();
		tris = PolyominoThing.readObjectsXML(WelltrisInterface.XML_FILE, null);
		if (tris == null) {
			System.err.println("Using fallback pieces.");
			polyominoThingFallback = new PolyominoThingFallback();
			tris = polyominoThingFallback.readStatic();
		}
		/*if (debug) {
			checkPolyomoinoes();
		}*/
	}
}

class Spec {
	volatile int units;
	volatile boolean diagonal, mixed, bonus;

	public Spec(int units,
			boolean diagonal, boolean mixed, boolean bonus) {
		this.units = units;
		this.diagonal = diagonal;
		this.mixed = mixed;
		this.bonus = bonus;
	}

	public Spec(int units) {
		this.units = units;
		this.diagonal = this.mixed = this.bonus = false;
	}

	void clone(Spec nextSpec) {
		units = nextSpec.units;
		diagonal = nextSpec.diagonal;
		mixed = nextSpec.mixed;
		bonus = nextSpec.bonus;
	}

	public void setDiagonal(boolean diagonal) {
		this.diagonal = diagonal;
	}

	public void setMixed(boolean mixed) {
		this.mixed = mixed;
	}

	public void setBonus(boolean bonus) {
		this.bonus = bonus;
	}

	public void setUnits(int units) {
		this.units = units;
	}
}

class Field {
	int pmid, cid; // pixmaps and colors

	public Field() {
		pmid = -1;
		cid = 0;
	}
}
